==> cirq-0.4.0.dev4220181119182308-py2-none-any.whl: cirq/sim/simulator_test.py <==
# coding=utf-8
# Copyright 2018 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
u"""Tests for simulator.py"""

from __future__ import with_statement
from __future__ import absolute_import
from typing import List, Dict

import numpy as np
import pytest

import cirq
from cirq.testing.mock import mock


@mock.patch.multiple(cirq.SimulatesSamples, _run=mock.Mock())
def test_run_simulator_run():
    simulator = cirq.SimulatesSamples()
    expected_measurements = {u'a': [[1]]}
    simulator._run.return_value = expected_measurements
    circuit = mock.Mock(cirq.Circuit)
    param_resolver = mock.Mock(cirq.ParamResolver)
    expected_result = cirq.TrialResult(repetitions=10,
                                       measurements=expected_measurements,
                                       params=param_resolver)
    assert expected_result == simulator.run(circuit=circuit,
                                            repetitions=10,
                                            param_resolver=param_resolver)
    simulator._run.assert_called_once_with(circuit=circuit,
                                           repetitions=10,
                                           param_resolver=param_resolver)


@mock.patch.multiple(cirq.SimulatesSamples, _run=mock.Mock())
def test_run_simulator_sweeps():
    simulator = cirq.SimulatesSamples()
    expected_measurements = {u'a': [[1]]}
    simulator._run.return_value = expected_measurements
    circuit = mock.Mock(cirq.Circuit)
    param_resolvers = [mock.Mock(cirq.ParamResolver),
                       mock.Mock(cirq.ParamResolver)]
    expected_results = [cirq.TrialResult(repetitions=10,
                                         measurements=expected_measurements,
                                         params=param_resolvers[0]),
                        cirq.TrialResult(repetitions=10,
                                         measurements=expected_measurements,
                                         params=param_resolvers[1])]
    assert expected_results == simulator.run_sweep(program=circuit,
                                                repetitions=10,
                                                params=param_resolvers)
    simulator._run.assert_called_with(circuit=circuit,
                                      repetitions=10,
                                      param_resolver=mock.ANY)
    assert simulator._run.call_count == 2


@mock.patch.multiple(cirq.SimulatesIntermediateWaveFunction,
                     _simulator_iterator=mock.Mock())
def test_wave_simulator():
    simulator = cirq.SimulatesIntermediateWaveFunction()

    final_state = np.array([1, 0, 0, 0])
    def steps(*args, **kwargs):
        result = mock.Mock()
        result.measurements = {u'a': [True, True]}
        yield result
        result = mock.Mock()
        result.measurements = {u'b': [True, False]}
        result.state.return_value = final_state
        yield result

    simulator._simulator_iterator.side_effect = steps
    circuit = mock.Mock(cirq.Circuit)
    param_resolver = mock.Mock(cirq.ParamResolver)
    qubit_order = mock.Mock(cirq.QubitOrder)
    result = simulator.simulate(circuit=circuit,
                                param_resolver=param_resolver,
                                qubit_order=qubit_order,
                                initial_state=2)
    np.testing.assert_equal(result.measurements[u'a'], [True, True])
    np.testing.assert_equal(result.measurements[u'b'], [True, False])
    assert set(result.measurements.keys()) == set([u'a', u'b'])
    assert result.params == param_resolver
    np.testing.assert_equal(result.final_state, final_state)


@mock.patch.multiple(cirq.SimulatesIntermediateWaveFunction,
                     _simulator_iterator=mock.Mock())
def test_wave_simulator_no_steps():
    simulator = cirq.SimulatesIntermediateWaveFunction()

    initial_state = np.array([1, 0, 0, 0], dtype=np.complex64)

    simulator._simulator_iterator.return_value = iter([])
    circuit = cirq.testing.random_circuit(2, 20, 0.99)
    param_resolver = mock.Mock(cirq.ParamResolver)
    qubit_order = circuit.all_qubits()
    result = simulator.simulate(circuit=circuit,
                                param_resolver=param_resolver,
                                qubit_order=list(qubit_order),
                                initial_state=initial_state)
    assert len(result.measurements) == 0
    assert result.params == param_resolver
    np.testing.assert_equal(result.final_state, initial_state)


@mock.patch.multiple(cirq.SimulatesIntermediateWaveFunction,
                     _simulator_iterator=mock.Mock())
def test_wave_simulator_sweeps():
    simulator = cirq.SimulatesIntermediateWaveFunction()

    final_state = np.array([1, 0, 0, 0])
    def steps(*args, **kwargs):
        result = mock.Mock()
        result.measurements = {u'a': np.array([True, True])}
        result.state.return_value = final_state
        yield result

    simulator._simulator_iterator.side_effect = steps
    circuit = mock.Mock(cirq.Circuit)
    param_resolvers = [mock.Mock(cirq.ParamResolver),
                       mock.Mock(cirq.ParamResolver)]
    qubit_order = mock.Mock(cirq.QubitOrder)
    results = simulator.simulate_sweep(program=circuit,
                                       params=param_resolvers,
                                       qubit_order=qubit_order,
                                       initial_state=2)
    expected_results = [
        cirq.SimulationTrialResult(
            measurements={u'a': np.array([True, True])},
            params=param_resolvers[0],
            final_state=final_state),
        cirq.SimulationTrialResult(
            measurements={u'a': np.array([True, True])},
            params=param_resolvers[1],
            final_state=final_state)
    ]
    assert results == expected_results


# Python 2 gives a different repr due to unicode strings being prefixed with u.
@cirq.testing.only_test_in_python3
def test_simulator_simulate_trial_result_repr():
    v = cirq.SimulationTrialResult(
        params=cirq.ParamResolver({u'a': 2}),
        measurements={u'm': np.array([1, 2])},
        final_state=np.array([0, 1, 0, 0]))

    assert repr(v) == (u"SimulationTrialResult("
                       u"params=cirq.ParamResolver({'a': 2}), "
                       u"measurements={'m': array([1, 2])}, "
                       u"final_state=array([0, 1, 0, 0]))")


def test_simulator_trial_result_equality():
    eq = cirq.testing.EqualsTester()
    eq.add_equality_group(
        cirq.SimulationTrialResult(
            params=cirq.ParamResolver({u'a': 2}),
            measurements={u'm': np.array([1, 0])},
            final_state=np.array([0, 1, 0, 0])))
    eq.add_equality_group(
        cirq.SimulationTrialResult(
            params=cirq.ParamResolver({u'a': 2}),
            measurements={u'm': np.array([1, 0])},
            final_state=np.array([0, 0, 1, 0])))
    eq.add_equality_group(
        cirq.SimulationTrialResult(
            params=cirq.ParamResolver({u'a': 3}),
            measurements={u'm': np.array([1, 0])},
            final_state=np.array([0, 0, 1, 0])))
    eq.add_equality_group(
        cirq.SimulationTrialResult(
            params=cirq.ParamResolver({u'a': 3}),
            measurements={u'm': np.array([0, 1])},
            final_state=np.array([0, 0, 1, 0])))


def test_simulator_trial_pretty_state():
    result = cirq.SimulationTrialResult(
        params=cirq.ParamResolver({u'a': 2}),
        measurements={u'm': np.array([1, 2])},
        final_state=np.array([0, 1, 0, 0]))
    assert result.dirac_notation() == u'|01⟩'


def test_step_result_pretty_state():
    class BasicStepResult(cirq.StepResult):

        def __init__(self, qubit_map,
                measurements):
            super(BasicStepResult, self).__init__(qubit_map, measurements)

        @property
        def state(self):
            return np.array([0, 1, 0, 0])

    step_result = BasicStepResult({}, {})
    assert step_result.dirac_notation() == u'|01⟩'


class FakeStepResult(cirq.StepResult):

    def __init__(self, ones_qubits):
        self._ones_qubits = set(ones_qubits)

    def state(self):
        pass

    def __setstate__(self, state):
        pass

    def sample(self, qubits, repetitions):
        return [[qubit in self._ones_qubits for qubit in qubits]] * repetitions


def test_step_sample_measurement_ops():
    q0, q1, q2 = cirq.LineQubit.range(3)
    measurement_ops = [cirq.measure(q0, q1), cirq.measure(q2)]
    step_result = FakeStepResult([q1])

    measurements = step_result.sample_measurement_ops(measurement_ops)
    np.testing.assert_equal(measurements,
                            {u'0,1': [[False, True]], u'2': [[False]]})


def test_step_sample_measurement_ops_repetitions():
    q0, q1, q2 = cirq.LineQubit.range(3)
    measurement_ops = [cirq.measure(q0, q1), cirq.measure(q2)]
    step_result = FakeStepResult([q1])

    measurements = step_result.sample_measurement_ops(measurement_ops,
                                                      repetitions=3)
    np.testing.assert_equal(measurements,
                            {u'0,1': [[False, True]] * 3, u'2': [[False]] * 3})


def test_step_sample_measurement_ops_no_measurements():
    step_result = FakeStepResult([])

    measurements = step_result.sample_measurement_ops([])
    assert measurements == {}


def test_step_sample_measurement_ops_not_measurement():
    q0 = cirq.LineQubit(0)
    step_result = FakeStepResult([q0])
    with pytest.raises(ValueError, match=u'MeasurementGate'):
        step_result.sample_measurement_ops([cirq.X(q0)])


def test_step_sample_measurement_ops_repeated_qubit():
    q0, q1, q2 = cirq.LineQubit.range(3)
    step_result = FakeStepResult([q0])
    with pytest.raises(ValueError, match=u'MeasurementGate'):
        step_result.sample_measurement_ops(
                [cirq.measure(q0), cirq.measure(q1, q2), cirq.measure(q0)])


class MultiHTestGate(cirq.Gate):
    def _decompose_(self, qubits):
        return cirq.H.on_each(qubits)


def test_simulates_composite():
    c = cirq.Circuit.from_ops(MultiHTestGate().on(*cirq.LineQubit.range(2)))
    expected = np.array([0.5] * 4)
    np.testing.assert_allclose(c.apply_unitary_effect_to_state(),
                               expected)
    np.testing.assert_allclose(cirq.Simulator().simulate(c).final_state,
                               expected)


def test_simulate_measurement_inversions():
    q = cirq.NamedQubit(u'q')

    c = cirq.Circuit.from_ops(cirq.MeasurementGate(key=u'q',
                                                   invert_mask=(True,)).on(q))
    assert cirq.Simulator().simulate(c).measurements == {u'q': np.array([True])}

    c = cirq.Circuit.from_ops(cirq.MeasurementGate(key=u'q',
                                                   invert_mask=(False,)).on(q))
    assert cirq.Simulator().simulate(c).measurements == {u'q': np.array([False])}

==> cirq-0.5.0-py2-none-any.whl: cirq/sim/simulator_test.py <==
# coding=utf-8
# Copyright 2018 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
u"""Tests for simulator.py"""

from __future__ import with_statement
from __future__ import absolute_import
import numpy as np
import pytest

import cirq
from cirq.testing.mock import mock


@mock.patch.multiple(cirq.SimulatesSamples,
                     __abstractmethods__=set(),
                     _run=mock.Mock())
def test_run_simulator_run():
    simulator = cirq.SimulatesSamples()
    expected_measurements = {u'a': np.array([[1]])}
    simulator._run.return_value = expected_measurements
    circuit = mock.Mock(cirq.Circuit)
    param_resolver = mock.Mock(cirq.ParamResolver)
    expected_result = cirq.TrialResult(repetitions=10,
                                       measurements=expected_measurements,
                                       params=param_resolver)
    assert expected_result == simulator.run(program=circuit,
                                            repetitions=10,
                                            param_resolver=param_resolver)
    simulator._run.assert_called_once_with(circuit=circuit,
                                           repetitions=10,
                                           param_resolver=param_resolver)


@mock.patch.multiple(cirq.SimulatesSamples,
                     __abstractmethods__=set(),
                     _run=mock.Mock())
def test_run_simulator_sweeps():
    simulator = cirq.SimulatesSamples()
    expected_measurements = {u'a': np.array([[1]])}
    simulator._run.return_value = expected_measurements
    circuit = mock.Mock(cirq.Circuit)
    param_resolvers = [mock.Mock(cirq.ParamResolver),
                       mock.Mock(cirq.ParamResolver)]
    expected_results = [cirq.TrialResult(repetitions=10,
                                         measurements=expected_measurements,
                                         params=param_resolvers[0]),
                        cirq.TrialResult(repetitions=10,
                                         measurements=expected_measurements,
                                         params=param_resolvers[1])]
    assert expected_results == simulator.run_sweep(program=circuit,
                                                   repetitions=10,
                                                   params=param_resolvers)
    simulator._run.assert_called_with(circuit=circuit,
                                      repetitions=10,
                                      param_resolver=mock.ANY)
    assert simulator._run.call_count == 2


@mock.patch.multiple(cirq.SimulatesIntermediateState,
                     __abstractmethods__=set(),
                     _simulator_iterator=mock.Mock())
def test_intermediate_simulator():
    simulator = cirq.SimulatesIntermediateState()

    final_simulator_state = np.array([1, 0, 0, 0])
    def steps(*args, **kwargs):
        result = mock.Mock()
        result.measurements = {u'a': [True, True]}
        yield result
        result = mock.Mock()
        result.measurements = {u'b': [True, False]}
        result.simulator_state.return_value = final_simulator_state
        yield result

    simulator._simulator_iterator.side_effect = steps
    circuit = mock.Mock(cirq.Circuit)
    param_resolver = mock.Mock(cirq.ParamResolver)
    qubit_order = mock.Mock(cirq.QubitOrder)
    result = simulator.simulate(program=circuit,
                                param_resolver=param_resolver,
                                qubit_order=qubit_order,
                                initial_state=2)
    np.testing.assert_equal(result.measurements[u'a'], [True, True])
    np.testing.assert_equal(result.measurements[u'b'], [True, False])
    assert set(result.measurements.keys()) == set([u'a', u'b'])
    assert result.params == param_resolver
    np.testing.assert_equal(result.final_simulator_state, final_simulator_state)


@mock.patch.multiple(cirq.SimulatesIntermediateState,
                     __abstractmethods__=set(),
                     _simulator_iterator=mock.Mock())
def test_intermediate_sweeps():
    simulator = cirq.SimulatesIntermediateState()

    final_state = np.array([1, 0, 0, 0])
    def steps(*args, **kwargs):
        result = mock.Mock()
        result.measurements = {u'a': np.array([True, True])}
        result.simulator_state.return_value = final_state
        yield result

    simulator._simulator_iterator.side_effect = steps
    circuit = mock.Mock(cirq.Circuit)
    param_resolvers = [mock.Mock(cirq.ParamResolver),
                       mock.Mock(cirq.ParamResolver)]
    qubit_order = mock.Mock(cirq.QubitOrder)
    results = simulator.simulate_sweep(program=circuit,
                                       params=param_resolvers,
                                       qubit_order=qubit_order,
                                       initial_state=2)
    expected_results = [
        cirq.SimulationTrialResult(
            measurements={u'a': np.array([True, True])},
            params=param_resolvers[0],
            final_simulator_state=final_state),
        cirq.SimulationTrialResult(
            measurements={u'a': np.array([True, True])},
            params=param_resolvers[1],
            final_simulator_state=final_state)
    ]
    assert results == expected_results


class FakeStepResult(cirq.StepResult):

    def __init__(self, ones_qubits):
        self._ones_qubits = set(ones_qubits)

    def simulator_state(self):
        pass

    def state_vector(self):
        pass

    def __setstate__(self, state):
        pass

    def sample(self, qubits, repetitions):
        return [[qubit in self._ones_qubits for qubit in qubits]] * repetitions


def test_step_sample_measurement_ops():
    q0, q1, q2 = cirq.LineQubit.range(3)
    measurement_ops = [cirq.measure(q0, q1), cirq.measure(q2)]
    step_result = FakeStepResult([q1])

    measurements = step_result.sample_measurement_ops(measurement_ops)
    np.testing.assert_equal(measurements,
                            {u'0,1': [[False, True]], u'2': [[False]]})


def test_step_sample_measurement_ops_repetitions():
    q0, q1, q2 = cirq.LineQubit.range(3)
    measurement_ops = [cirq.measure(q0, q1), cirq.measure(q2)]
    step_result = FakeStepResult([q1])

    measurements = step_result.sample_measurement_ops(measurement_ops,
                                                      repetitions=3)
    np.testing.assert_equal(measurements,
                            {u'0,1': [[False, True]] * 3, u'2': [[False]] * 3})


def test_step_sample_measurement_ops_no_measurements():
    step_result = FakeStepResult([])

    measurements = step_result.sample_measurement_ops([])
    assert measurements == {}


def test_step_sample_measurement_ops_not_measurement():
    q0 = cirq.LineQubit(0)
    step_result = FakeStepResult([q0])
    with pytest.raises(ValueError, match=u'MeasurementGate'):
        step_result.sample_measurement_ops([cirq.X(q0)])


def test_step_sample_measurement_ops_repeated_qubit():
    q0, q1, q2 = cirq.LineQubit.range(3)
    step_result = FakeStepResult([q0])
    with pytest.raises(ValueError, match=u'MeasurementGate'):
        step_result.sample_measurement_ops(
                [cirq.measure(q0), cirq.measure(q1, q2), cirq.measure(q0)])


def test_simulation_trial_result_equality():
    eq = cirq.testing.EqualsTester()
    eq.add_equality_group(
        cirq.SimulationTrialResult(params=cirq.ParamResolver({}),
                                   measurements={},
                                   final_simulator_state=()),
        cirq.SimulationTrialResult(params=cirq.ParamResolver({}),
                                   measurements={},
                                   final_simulator_state=()))
    eq.add_equality_group(
        cirq.SimulationTrialResult(params=cirq.ParamResolver({u's': 1}),
                                   measurements={},
                                   final_simulator_state=()))
    eq.add_equality_group(
        cirq.SimulationTrialResult(params=cirq.ParamResolver({u's': 1}),
                                   measurements={u'm': np.array([[1]])},
                                   final_simulator_state=()))
    eq.add_equality_group(
        cirq.SimulationTrialResult(params=cirq.ParamResolver({u's': 1}),
                                   measurements={u'm': np.array([[1]])},
                                   final_simulator_state=(0, 1)))


# Python 2 gives a different repr due to unicode strings being prefixed with u.
@cirq.testing.only_test_in_python3
def test_simulation_trial_result_repr():
    assert repr(cirq.SimulationTrialResult(params=cirq.ParamResolver({u's': 1}),
                                           measurements={u'm': np.array([[1]])},
                                           final_simulator_state=(0, 1))) == (
               u"cirq.SimulationTrialResult("
               u"params=cirq.ParamResolver({'s': 1}), "
               u"measurements={'m': array([[1]])}, "
               u"final_simulator_state=(0, 1))")


def test_simulation_trial_result_str():
    assert unicode(
        cirq.SimulationTrialResult(
            params=cirq.ParamResolver({u's': 1}),
            measurements={},
            final_simulator_state=(0, 1))) == u'(no measurements)'

    assert unicode(cirq.SimulationTrialResult(
        params=cirq.ParamResolver({u's': 1}),
        measurements={u'm': np.array([[1]])},
        final_simulator_state=(0, 1))) == u'm=1'


def test_pretty_print():
    result = cirq.SimulationTrialResult(cirq.ParamResolver(), {}, np.array([1]))

    # Test Jupyter console output from
    class FakePrinter(object):

        def __init__(self):
            self.text_pretty = u''

        def text(self, to_print):
            self.text_pretty += to_print

    p = FakePrinter()
    result._repr_pretty_(p, False)
    assert p.text_pretty == u'(no measurements)'

    # Test cycle handling
    p = FakePrinter()
    result._repr_pretty_(p, True)
    assert p.text_pretty == u'SimulationTrialResult(...)'
